import aiohttp
import asyncio
from botstory.ast import story_context
from botstory.integrations.commonhttp import errors as commonhttp_errors
from botstory.utils import answer
import logging
import unittest
from unittest import mock
import pytest

from . import messenger
from .. import commonhttp, mockdb, mockhttp
from ... import di, Story, utils
from ...middlewares import any, option, sticker

logger = logging.getLogger(__name__)

story = None


def teardown_function(function):
    logger.debug('tear down!')
    story.clear()


@pytest.mark.asyncio
async def test_send_text_message():
    user = utils.build_fake_user()

    global story
    story = Story()

    interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.start()

    await interface.send_text_message(
        recipient=user, text='hi!', quick_replies=None
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty1',
        },
        json={
            'message': {
                'text': 'hi!',
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_truncate_long_message():
    user = utils.build_fake_user()

    global story
    story = Story()

    interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.start()

    very_long_message = 'very_long_message' * 100
    await interface.send_text_message(
        recipient=user,
        text=very_long_message,
        quick_replies=None,
        options={
            'overflow': 'cut'
        }
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty1',
        },
        json={
            'message': {
                'text': very_long_message[:640],
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_truncate_with_ellipsis_long_message_by_default():
    user = utils.build_fake_user()

    global story
    story = Story()

    interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.start()

    very_long_message = 'very_long_message' * 100
    await interface.send_text_message(
        recipient=user,
        text=very_long_message,
        quick_replies=None,
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty1',
        },
        json={
            'message': {
                'text': very_long_message[:638] + '\u2026',
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_send_list():
    with answer.Talk() as talk:
        story = talk.story
        fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
        mock_http = story.use(mockhttp.MockHttpInterface())
        await story.start()
        await fb_interface.send_list(
            recipient=talk.user,
            elements=[{
                'title': 'Classic T-Shirt Collection',  # (*) required
                'image_url': 'https://peterssendreceiveapp.ngrok.io/img/collection.png',
                'subtitle': 'See all our colors',
                'default_action': {
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/shop_collection',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                },
                'buttons': [{
                    'title': 'View',
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/collection',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                }]
            }, {
                'title': 'Classic White T-Shirt',
                'image_url': 'https://peterssendreceiveapp.ngrok.io/img/white-t-shirt.png',
                'subtitle': '100% Cotton, 200% Comfortable',
                'default_action': {
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/view?item=100',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                },
                'buttons': [{
                    'title': 'Shop Now',
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/shop?item=100',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                }]
            }, {
                'title': 'Classic Blue T-Shirt',
                'image_url': 'https://peterssendreceiveapp.ngrok.io/img/blue-t-shirt.png',
                'subtitle': '100% Cotton, 200% Comfortable',
                'default_action': {
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/view?item=101',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                },
                'buttons': [{
                    'title': 'Shop Now',
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/shop?item=101',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                }]
            }, {
                'title': 'Classic Black T-Shirt',
                'image_url': 'https://peterssendreceiveapp.ngrok.io/img/black-t-shirt.png',
                'subtitle': '100% Cotton, 200% Comfortable',
                'default_action': {
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/view?item=102',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                },
                'buttons': [{
                    'title': 'Shop Now',
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/shop?item=102',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                }]
            }], buttons=[{
                'title': 'View More',
                'payload': 'payload',
            }])
        mock_http.post.assert_called_with(
            'https://graph.facebook.com/v2.6/me/messages/',
            params={
                'access_token': 'qwerty1',
            },
            json={
                'message': {
                    'attachment': {
                        'type': 'template',
                        'payload': {
                            'template_type': 'list',
                            'top_element_style': 'large',
                            'elements': [{
                                'title': 'Classic T-Shirt Collection',  # (*) required
                                'image_url': 'https://peterssendreceiveapp.ngrok.io/img/collection.png',
                                'subtitle': 'See all our colors',
                                'default_action': {
                                    'type': 'web_url',
                                    'url': 'https://peterssendreceiveapp.ngrok.io/shop_collection',
                                    'messenger_extensions': True,
                                    'webview_height_ratio': 'tall',
                                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                                },
                                'buttons': [{
                                    'title': 'View',
                                    'type': 'web_url',
                                    'url': 'https://peterssendreceiveapp.ngrok.io/collection',
                                    'messenger_extensions': True,
                                    'webview_height_ratio': 'tall',
                                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                                }]
                            }, {
                                'title': 'Classic White T-Shirt',
                                'image_url': 'https://peterssendreceiveapp.ngrok.io/img/white-t-shirt.png',
                                'subtitle': '100% Cotton, 200% Comfortable',
                                'default_action': {
                                    'type': 'web_url',
                                    'url': 'https://peterssendreceiveapp.ngrok.io/view?item=100',
                                    'messenger_extensions': True,
                                    'webview_height_ratio': 'tall',
                                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                                },
                                'buttons': [{
                                    'title': 'Shop Now',
                                    'type': 'web_url',
                                    'url': 'https://peterssendreceiveapp.ngrok.io/shop?item=100',
                                    'messenger_extensions': True,
                                    'webview_height_ratio': 'tall',
                                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                                }]
                            }, {
                                'title': 'Classic Blue T-Shirt',
                                'image_url': 'https://peterssendreceiveapp.ngrok.io/img/blue-t-shirt.png',
                                'subtitle': '100% Cotton, 200% Comfortable',
                                'default_action': {
                                    'type': 'web_url',
                                    'url': 'https://peterssendreceiveapp.ngrok.io/view?item=101',
                                    'messenger_extensions': True,
                                    'webview_height_ratio': 'tall',
                                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                                },
                                'buttons': [{
                                    'title': 'Shop Now',
                                    'type': 'web_url',
                                    'url': 'https://peterssendreceiveapp.ngrok.io/shop?item=101',
                                    'messenger_extensions': True,
                                    'webview_height_ratio': 'tall',
                                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                                }]
                            }, {
                                'title': 'Classic Black T-Shirt',
                                'image_url': 'https://peterssendreceiveapp.ngrok.io/img/black-t-shirt.png',
                                'subtitle': '100% Cotton, 200% Comfortable',
                                'default_action': {
                                    'type': 'web_url',
                                    'url': 'https://peterssendreceiveapp.ngrok.io/view?item=102',
                                    'messenger_extensions': True,
                                    'webview_height_ratio': 'tall',
                                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                                },
                                'buttons': [{
                                    'title': 'Shop Now',
                                    'type': 'web_url',
                                    'url': 'https://peterssendreceiveapp.ngrok.io/shop?item=102',
                                    'messenger_extensions': True,
                                    'webview_height_ratio': 'tall',
                                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                                }]
                            }],
                            'buttons': [
                                {
                                    'title': 'View More',
                                    'type': 'postback',
                                    'payload': 'payload'
                                }
                            ]
                        }
                    }
                },
                'recipient': {
                    'id': talk.user['facebook_user_id'],
                },
            }
        )


@pytest.mark.asyncio
async def test_should_send_template_based_message():
    with answer.Talk() as talk:
        story = talk.story
        fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
        mock_http = story.use(mockhttp.MockHttpInterface())
        await story.start()
        payload = {
            'template_type': 'receipt',
            'recipient_name': 'Stephane Crozatier',
            'order_number': '12345678902',
            'currency': 'USD',
            'payment_method': 'Visa 2345',
            'order_url': 'http://petersapparel.parseapp.com/order?order_id=123456',
            'timestamp': '1428444852',
            'elements': [{
                'title': 'Classic White T-Shirt',
                'subtitle': '100% Soft and Luxurious Cotton',
                'quantity': 2,
                'price': 50,
                'currency': 'USD',
                'image_url': 'http://petersapparel.parseapp.com/img/whiteshirt.png'
            }, {
                'title': 'Classic Gray T-Shirt',
                'subtitle': '100% Soft and Luxurious Cotton',
                'quantity': 1,
                'price': 25,
                'currency': 'USD',
                'image_url': 'http://petersapparel.parseapp.com/img/grayshirt.png'
            }],
            'address': {
                'street_1': '1 Hacker Way',
                'street_2': '',
                'city': 'Menlo Park',
                'postal_code': '94025',
                'state': 'CA',
                'country': 'US'
            },
            'summary': {
                'subtotal': 75.00,
                'shipping_cost': 4.95,
                'total_tax': 6.19,
                'total_cost': 56.14
            },
            'adjustments': [{
                'name': 'New Customer Discount',
                'amount': 20
            }, {
                'name': '$10 Off Coupon',
                'amount': 10
            }]
        }
        await fb_interface.send_template(talk.user, payload)
        mock_http.post.assert_called_with(
            'https://graph.facebook.com/v2.6/me/messages/',
            params={
                'access_token': 'qwerty1',
            },
            json={
                'message': {
                    'attachment': {
                        'type': 'template',
                        'payload': payload,
                    }
                },
                'recipient': {
                    'id': talk.user['facebook_user_id'],
                },
            }
        )


@pytest.mark.asyncio
async def test_send_image():
    with answer.Talk() as talk:
        story = talk.story
        fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
        mock_http = story.use(mockhttp.MockHttpInterface())
        await story.start()
        await fb_interface.send_image(talk.user, 'http://shevchenko.ua/image.gif')
        mock_http.post.assert_called_with(
            'https://graph.facebook.com/v2.6/me/messages/',
            params={
                'access_token': 'qwerty1',
            },
            json={
                'message': {
                    'attachment': {
                        'type': 'image',
                        'payload': {
                            'url': 'http://shevchenko.ua/image.gif',
                        },
                    }
                },
                'recipient': {
                    'id': talk.user['facebook_user_id'],
                },
            }
        )


def should_post_attachment(mock_http, talk):
    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty1',
        },
        json={
            'message': {
                'attachment': {
                    'type': 'image',
                    'payload': {
                        'url': 'http://shevchenko.ua/image.gif',
                    },
                }
            },
            'recipient': {
                'id': talk.user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_retry_send_image():
    with answer.Talk() as talk:
        story = talk.story
        fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
        mock_http = story.use(mockhttp.MockHttpInterface(
            post_raise=commonhttp_errors.HttpRequestError(),
        ))
        await story.start()

        send_task = fb_interface.send_image(talk.user, 'http://shevchenko.ua/image.gif', options={
            'retry_times': 3,
            'retry_delay': 0.1,
        })

        async def lazy_fix_http():
            # here should pass first 2 retry
            await asyncio.sleep(0.15)
            # than we change mock http without post raise
            # so on 3 try it should pass without problem
            story.use(mockhttp.MockHttpInterface())

        await asyncio.gather(
            lazy_fix_http(),
            send_task,
        )

        should_post_attachment(mock_http, talk)


@pytest.mark.asyncio
async def test_retry_send_image_should_fail_on_tries_exceed():
    with answer.Talk() as talk:
        story = talk.story
        fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
        mock_http = story.use(mockhttp.MockHttpInterface(
            post_raise=commonhttp_errors.HttpRequestError(),
        ))
        await story.start()

        with pytest.raises(commonhttp_errors.HttpRequestError):
            await fb_interface.send_image(talk.user, 'http://shevchenko.ua/image.gif', options={
                'retry_times': 3,
                'retry_delay': 0.1,
            })

        should_post_attachment(mock_http, talk)


@pytest.mark.asyncio
async def test_integration():
    user = utils.build_fake_user()

    global story
    story = Story()

    story.use(messenger.FBInterface(page_access_token='qwerty2'))
    story.use(mockdb.MockDB())
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.say('hi there!', user=user)

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty2',
        },
        json={
            'message': {
                'text': 'hi there!',
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_quick_replies():
    user = utils.build_fake_user()

    global story
    story = Story()

    story.use(messenger.FBInterface(page_access_token='qwerty3'))
    story.use(mockdb.MockDB())
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.ask(
        'Which color do you like?',
        quick_replies=[{
            'title': 'Red',
            'payload': 0xff0000,
        }, {
            'title': 'Green',
            'payload': 0x00ff00,
        }, {
            'title': 'Blue',
            'payload': 0x0000ff,
        }],
        user=user,
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty3',
        },
        json={
            'message': {
                'text': 'Which color do you like?',
                'quick_replies': [
                    {
                        'content_type': 'text',
                        'title': 'Red',
                        'payload': 0xff0000,
                    },
                    {
                        'content_type': 'text',
                        'title': 'Green',
                        'payload': 0x00ff00,
                    },
                    {
                        'content_type': 'text',
                        'title': 'Blue',
                        'payload': 0x0000ff,
                    },
                ],
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_quick_replies_with_location():
    user = utils.build_fake_user()

    global story
    story = Story()

    story.use(messenger.FBInterface(page_access_token='qwerty3'))
    story.use(mockdb.MockDB())
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.ask(
        'Where do you live?',
        quick_replies=[{
            'content_type': 'location',
        }, {
            'title': 'Europe',
            'payload': 'SET_LOCATION_EU',
        }, {
            'title': 'US :',
            'payload': 'SET_LOCATION_US',
        }, {
            'title': 'Ukraine',
            'payload': 'SET_LOCATION_UA',
        }, ],
        user=user,
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty3',
        },
        json={
            'message': {
                'text': 'Where do you live?',
                'quick_replies': [
                    {
                        'content_type': 'location',
                    }, {
                        'content_type': 'text',
                        'title': 'Europe',
                        'payload': 'SET_LOCATION_EU',
                    }, {
                        'content_type': 'text',
                        'title': 'US :',
                        'payload': 'SET_LOCATION_US',
                    }, {
                        'content_type': 'text',
                        'title': 'Ukraine',
                        'payload': 'SET_LOCATION_UA',
                    },
                ],
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_setup_webhook():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.start()

    mock_http.webhook.assert_called_with(
        '/webhook',
        fb_interface.handle,
        'some-token',
    )


@pytest.mark.asyncio
async def test_should_request_user_data_once_we_do_not_know_current_user():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(
        page_access_token='qwerty4',
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    http = story.use(mockhttp.MockHttpInterface(get={
        'first_name': 'Peter',
        'last_name': 'Chang',
        'profile_pic': 'https://fbcdn-profile-a.akamaihd.net/hprofile-ak-xpf1/v/t1.0-1/p200x200/13055603_10105219398495383_8237637584159975445_n.jpg?oh=1d241d4b6d4dac50eaf9bb73288ea192&oe=57AF5C03&__gda__=1470213755_ab17c8c8e3a0a447fed3f272fa2179ce',
        'locale': 'en_US',
        'timezone': -7,
        'gender': 'male'
    }))
    story.use(mockdb.MockDB())

    await fb_interface.process({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    http.get.assert_called_with(
        'https://graph.facebook.com/v2.6/USER_ID',
        params={
            'access_token': 'qwerty4',
        },
    )


@pytest.mark.asyncio
async def test_should_request_user_data_and_fail():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(
        page_access_token='qwerty5',
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    story.use(mockhttp.MockHttpInterface(
        get_raise=commonhttp.errors.HttpRequestError()))
    db = story.use(mockdb.MockDB())

    await fb_interface.process({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    assert (await db.get_user(facebook_user_id='USER_ID')).no_fb_profile is True


@pytest.mark.asyncio
async def test_webhook_handler_should_return_ok_status_if_http_fail():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(
        page_access_token='qwerty6',
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    story.use(mockhttp.MockHttpInterface(get_raise=commonhttp.errors.HttpRequestError()))
    story.use(mockdb.MockDB())

    res = await fb_interface.process({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    assert res['status'] == 200


@pytest.mark.asyncio
async def test_webhook_handler_should_return_ok_status_in_any_case():
    global story
    story = Story()

    fb_interface = messenger.FBInterface()
    with mock.patch('botstory.integrations.fb.messenger.logger') as mock_logger:
        res = await fb_interface.process({
            'object': 'page',
            'entry': [{
                'id': 'PAGE_ID',
                'time': 1473204787206,
                'messaging': [
                    {
                        'sender': {
                            'id': 'USER_ID'
                        },
                        'recipient': {
                            'id': 'PAGE_ID'
                        },
                        'timestamp': 1458692752478,
                        'message': {
                            'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                            'seq': 73,
                            'text': 'hello, world!'
                        }
                    }
                ]
            }]
        })

        assert mock_logger.debug.calledWith()

    assert res['status'] == 200


# integration

@pytest.fixture
def build_fb_interface():
    async def builder():
        user = utils.build_fake_user()
        session = utils.build_fake_session()

        global story
        story = Story()

        storage = story.use(mockdb.MockDB())
        fb = story.use(messenger.FBInterface(page_access_token='qwerty'))

        await story.start()

        await storage.set_session(session)
        await storage.set_user(user)

        return fb, story

    return builder


@pytest.mark.asyncio
async def test_handler_raw_text(build_fb_interface):
    fb_interface, story = await build_fb_interface()

    correct_trigger = utils.SimpleTrigger()
    incorrect_trigger = utils.SimpleTrigger()

    @story.on('hello, world!')
    def correct_story():
        @story.part()
        def store_result(ctx):
            correct_trigger.receive(story_context.get_message_data(ctx))

    @story.on('Goodbye, world!')
    def incorrect_story():
        @story.part()
        def store_result(ctx):
            incorrect_trigger.receive(story_context.get_message_data(ctx))

    await fb_interface.process({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    assert incorrect_trigger.value is None
    assert correct_trigger.value == {
        'text': {
            'raw': 'hello, world!'
        }
    }


@pytest.mark.asyncio
async def test_handler_selected_option(build_fb_interface):
    fb_interface, story = await build_fb_interface()

    correct_trigger = utils.SimpleTrigger()
    incorrect_trigger = utils.SimpleTrigger()

    @story.on(receive=option.Equal('GREEN'))
    def correct_story():
        @story.part()
        def store_result(ctx):
            correct_trigger.receive(story_context.get_message_data(ctx))

    @story.on(receive=option.Equal('BLUE'))
    def incorrect_story():
        @story.part()
        def store_result(ctx):
            incorrect_trigger.receive(story_context.get_message_data(ctx))

    await fb_interface.process({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [{
                'sender': {
                    'id': 'USER_ID'
                },
                'recipient': {
                    'id': 'PAGE_ID'
                },
                'timestamp': 1458692752478,
                'message': {
                    'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                    'seq': 73,
                    'text': 'Green!',
                    'quick_reply': {
                        'payload': 'GREEN'
                    }
                }
            }]
        }]
    })

    assert incorrect_trigger.value is None
    assert correct_trigger.value == {
        'option': {
            'value': 'GREEN',
        },
        'text': {
            'raw': 'Green!'
        }
    }


@pytest.mark.asyncio
async def test_handler_postback(build_fb_interface):
    fb_interface, story = await build_fb_interface()

    correct_trigger = utils.SimpleTrigger()
    incorrect_trigger = utils.SimpleTrigger()

    @story.on(receive=option.Equal('GREEN'))
    def correct_story():
        @story.part()
        def store_result(ctx):
            correct_trigger.receive(story_context.get_message_data(ctx))

    @story.on(receive=option.Equal('BLUE'))
    def incorrect_story():
        @story.part()
        def store_result(ctx):
            incorrect_trigger.receive(story_context.get_message_data(ctx))

    await fb_interface.process({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [{
                'sender': {
                    'id': 'USER_ID'
                },
                'recipient': {
                    'id': 'PAGE_ID'
                },
                'timestamp': 1458692752478,
                'postback': {
                    'payload': 'GREEN'
                },
            }]
        }]
    })

    assert incorrect_trigger.value is None
    assert correct_trigger.value == {
        'option': {'value': 'GREEN'},
    }


@pytest.mark.asyncio
async def test_handler_thumbsup(build_fb_interface):
    fb_interface, story = await build_fb_interface()

    like_is_here_trigger = utils.SimpleTrigger()

    @story.on(receive=sticker.Like())
    def like_story():
        @story.part()
        def store_result(ctx):
            like_is_here_trigger.passed()

    await fb_interface.process({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [{
                'sender': {
                    'id': 'USER_ID'
                },
                'recipient': {
                    'id': 'PAGE_ID'
                },
                'timestamp': 1458692752478,
                'message': {
                    'sticker_id': sticker.SMALL_LIKE,
                }
            }]
        }]
    })

    assert like_is_here_trigger.is_passed()


@pytest.mark.asyncio
async def test_should_not_process_echo_delivery_and_read_messages_as_regular(build_fb_interface):
    fb_interface, story = await build_fb_interface()

    echo_trigger = utils.SimpleTrigger()

    @story.on(receive=any.Any())
    def one_story():
        @story.part()
        def sync_part(message):
            echo_trigger.passed()

    await fb_interface.process({
        'entry': [
            {
                'id': '329188380752158',
                'messaging': [{
                    'message': {
                        'app_id': 345865645763384,
                        'is_echo': 'True',
                        'mid': 'mid.1477350590023:38b1efd593',
                        'seq': 323,
                        'text': 'Hm I dont know what is it'
                    },
                    'recipient': {
                        'id': '1034692249977067'
                    },
                    'sender': {
                        'id': '329188380752158'
                    },
                    'timestamp': 1477350590023
                }, {
                    'read': {
                        'seq': 2697,
                        'watermark': 1477354670744
                    },
                    'recipient': {
                        'id': '329188380752158'
                    },
                    'sender': {
                        'id': '1034692249977067'
                    },
                    'timestamp': 1477354672037
                }, {
                    'delivery': {
                        'mids': [
                            'mid.1477354667117:8fedc43d37'
                        ],
                        'seq': 2679,
                        'watermark': 1477354668538
                    },
                    'recipient': {
                        'id': '329188380752158'
                    },
                    'sender': {
                        'id': '1034692249977067'
                    },
                    'timestamp': 0
                }],
                'time': 1477350590772
            }
        ],
        'object': 'page'
    })

    assert not echo_trigger.is_triggered


@pytest.mark.asyncio
async def test_set_greeting_text():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty7'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.set_greeting_text('Hi there {{user_first_name}}!')

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty7',
        },
        json={
            'greeting': [{
                'locale': 'default',
                'text': 'Hi there {{user_first_name}}!',
            }],
        }
    )


@pytest.mark.asyncio
async def test_can_set_greeting_text_before_inject_http():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty8'))
    await fb_interface.set_greeting_text('Hi there {{user_first_name}}!')

    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.setup()

    # give few a moment for lazy initialization of greeting text
    await asyncio.sleep(0.1)

    mock_http.post.assert_has_calls([unittest.mock.call(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty8',
        },
        json={
            'greeting': [{
                'locale': 'default',
                'text': 'Hi there {{user_first_name}}!',
            }],
        }
    )])


@pytest.mark.asyncio
async def test_can_set_greeting_text_in_constructor():
    global story
    story = Story()

    fb = story.use(messenger.FBInterface(
        greeting_text='Hi there {{user_first_name}}!',
        page_access_token='qwerty9',
    ))

    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.setup()

    # give few a moment for lazy initialization of greeting text
    await asyncio.sleep(0.1)

    mock_http.delete.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty9',
        },
        json={
            'fields': [
                'greeting',
            ]
        },
    )

    mock_http.post.assert_has_calls([unittest.mock.call(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty9',
        },
        json={
            'greeting': [{
                'locale': 'default',
                'text': 'Hi there {{user_first_name}}!',
            }],
        }
    )])


@pytest.mark.asyncio
async def test_remove_greeting_text():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty10'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.remove_greeting_text()

    mock_http.delete.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty10',
        },
        json={
            'fields': [
                'greeting',
            ]
        }
    )


@pytest.mark.asyncio
async def test_set_greeting_call_to_action_payload():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty11'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.set_greeting_call_to_action_payload('SOME_PAYLOAD')

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty11',
        },
        json={
            'get_started': {'payload': 'SOME_PAYLOAD'}
        }
    )


@pytest.mark.asyncio
async def test_remove_greeting_call_to_action_payload():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty12'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.remove_greeting_call_to_action_payload()

    mock_http.delete.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty12',
        },
        json={
            'fields': [
                'get_started',
            ]
        }
    )


@pytest.mark.asyncio
async def test_set_persistent_menu():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty13'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.set_persistent_menu([{
        'type': 'postback',
        'title': 'Help',
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
    }, {
        'type': 'web_url',
        'title': 'View Website',
        'url': 'http://petersapparel.parseapp.com/'
    }])

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty13',
        },
        json={
            'persistent_menu': [
                {
                    'locale': 'default',
                    'call_to_actions': [{
                        'type': 'postback',
                        'title': 'Help',
                        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP',
                    }, {
                        'type': 'web_url',
                        'title': 'View Website',
                        'url': 'http://petersapparel.parseapp.com/',
                    }],
                },
            ],
        }
    )


@pytest.mark.asyncio
async def test_set_persistent_menu_with_locales():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty13'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.set_persistent_menu([
        {
            'locale': 'default',
            "composer_input_disabled": True,
            'call_to_actions': [{
                'type': 'postback',
                'title': 'Help',
                'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP',
            }, {
                'type': 'web_url',
                'title': 'View Website',
                'url': 'http://petersapparel.parseapp.com/',
            }],
        },
        {
            'locale': 'uk_UA',
            "composer_input_disabled": True,
            'call_to_actions': [{
                'type': 'postback',
                'title': 'Допомога',
                'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP',
            }, {
                'type': 'web_url',
                'title': 'Переглянути сторінку',
                'url': 'http://petersapparel.parseapp.com/',
            }],
        },
    ])

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty13',
        },
        json={
            'persistent_menu': [
                {
                    'locale': 'default',
                    "composer_input_disabled": True,
                    'call_to_actions': [{
                        'type': 'postback',
                        'title': 'Help',
                        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP',
                    }, {
                        'type': 'web_url',
                        'title': 'View Website',
                        'url': 'http://petersapparel.parseapp.com/',
                    }],
                },
                {
                    'locale': 'uk_UA',
                    "composer_input_disabled": True,
                    'call_to_actions': [{
                        'type': 'postback',
                        'title': 'Допомога',
                        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP',
                    }, {
                        'type': 'web_url',
                        'title': 'Переглянути сторінку',
                        'url': 'http://petersapparel.parseapp.com/',
                    }],
                },
            ],
        }
    )


@pytest.mark.asyncio
async def test_can_set_persistent_menu_before_http():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty14'))
    await fb_interface.set_persistent_menu([{
        'type': 'postback',
        'title': 'Help',
        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
    }, {
        'type': 'web_url',
        'title': 'View Website',
        'url': 'http://petersapparel.parseapp.com/'
    }])

    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.setup()

    # give few a moment for lazy initialization of greeting text
    await asyncio.sleep(0.1)

    mock_http.post.assert_has_calls([unittest.mock.call(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty14',
        },
        json={
            'persistent_menu': [
                {
                    'locale': 'default',
                    'call_to_actions': [{
                        'type': 'postback',
                        'title': 'Help',
                        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
                    }, {
                        'type': 'web_url',
                        'title': 'View Website',
                        'url': 'http://petersapparel.parseapp.com/'
                    }],
                },
            ],
        }
    )])


@pytest.mark.asyncio
async def test_can_set_persistent_menu_inside_of_constructor():
    global story
    story = Story()

    story.use(messenger.FBInterface(
        page_access_token='qwerty15',
        persistent_menu=[{
            'type': 'postback',
            'title': 'Help',
            'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
        }, {
            'type': 'web_url',
            'title': 'View Website',
            'url': 'http://petersapparel.parseapp.com/'
        }]
    ))

    mock_http = story.use(mockhttp.MockHttpInterface())

    await story.setup()

    # give few a moment for lazy initialization of greeting text
    await asyncio.sleep(0.1)

    mock_http.delete.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty15',
        },
        json={'fields': [
            'persistent_menu',
        ]}
    )

    mock_http.post.assert_has_calls([unittest.mock.call(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty15',
        },
        json={
            'persistent_menu': [
                {
                    'locale': 'default',
                    'call_to_actions': [{
                        'type': 'postback',
                        'title': 'Help',
                        'payload': 'DEVELOPER_DEFINED_PAYLOAD_FOR_HELP'
                    }, {
                        'type': 'web_url',
                        'title': 'View Website',
                        'url': 'http://petersapparel.parseapp.com/'
                    }],
                },
            ],
        }
    )], any_order=True)


@pytest.mark.asyncio
async def test_subscribe_to_page_on_setup():
    with answer.Talk() as talk:
        story = talk.story
        fb_interface = messenger.FBInterface(
            page_access_token='one-token',
        )
        fb_interface.subscribe = aiohttp.test_utils.make_mocked_coro()
        http_interface = mockhttp.MockHttpInterface()
        story.use(fb_interface)
        story.use(http_interface)

        await story.setup()
        fb_interface.subscribe.assert_called_with()


@pytest.mark.asyncio
async def test_remove_persistent_menu():
    global story
    story = Story()

    fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty16'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await fb_interface.remove_persistent_menu()

    mock_http.delete.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messenger_profile',
        params={
            'access_token': 'qwerty16',
        },
        json={'fields': [
            'persistent_menu',
        ]}
    )


def test_get_fb_as_deps():
    global story
    story = Story()

    story.use(messenger.FBInterface())

    with di.child_scope():
        @di.desc()
        class OneClass:
            @di.inject()
            def deps(self, fb):
                self.fb = fb

        assert isinstance(di.injector.get('one_class').fb, messenger.FBInterface)


def test_bind_fb_deps():
    global story
    story = Story()

    story.use(messenger.FBInterface())
    story.use(mockdb.MockDB())
    story.use(mockhttp.MockHttpInterface())

    with di.child_scope():
        @di.desc()
        class OneClass:
            @di.inject()
            def deps(self, fb):
                self.fb = fb

        assert isinstance(di.injector.get('one_class').fb.http, mockhttp.MockHttpInterface)
        assert isinstance(di.injector.get('one_class').fb.storage, mockdb.MockDB)


def one_message(talk):
    return {
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': talk.user['facebook_user_id'],
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    }


@pytest.mark.asyncio
async def test_quickly_returns_200ok():
    trigger = utils.SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story
        fb_interface = story.use(messenger.FBInterface(page_access_token='qwerty1'))
        story.use(mockdb.MockDB())
        story.use(mockhttp.MockHttpInterface())

        @story.on('hello, world!')
        def one_story():
            @story.part()
            def store_result(ctx):
                trigger.passed()

        await story.start()

        res = await fb_interface.handle(one_message(talk))
        assert res == {
            'status': 200,
            'text': 'Ok!',
        }
        assert not trigger.is_passed()
        await asyncio.sleep(0)
        assert trigger.is_passed()


@pytest.mark.asyncio
async def test_subscribe():
    fb_interface = messenger.FBInterface(
        page_access_token='one-token',
    )
    http_interface = mockhttp.MockHttpInterface()
    fb_interface.add_http(http_interface)
    await fb_interface.subscribe()
    http_interface.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/subscribed_apps',
        params={
            'access_token': 'one-token',
        },
    )


@pytest.mark.asyncio
async def test_start_typing():
    fake_user = utils.build_fake_user()

    fb_interface = messenger.FBInterface(
        page_access_token='one-token',
    )
    http_interface = mockhttp.MockHttpInterface()
    fb_interface.add_http(http_interface)

    await fb_interface.start_typing(fake_user)
    http_interface.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages',
        params={
            'access_token': 'one-token',
        },
        json={
            'recipient': {
                'id': fake_user['facebook_user_id'],
            },
            'sender_action': 'typing_on',
        }
    )


@pytest.mark.asyncio
async def test_stop_typing():
    fake_user = utils.build_fake_user()

    fb_interface = messenger.FBInterface(
        page_access_token='one-token',
    )
    http_interface = mockhttp.MockHttpInterface()
    fb_interface.add_http(http_interface)

    await fb_interface.stop_typing(fake_user)
    http_interface.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages',
        params={
            'access_token': 'one-token',
        },
        json={
            'recipient': {
                'id': fake_user['facebook_user_id'],
            },
            'sender_action': 'typing_off',
        }
    )



import aiohttp
import logging
import pytest

from botstory import Story
import botstory.integrations.fb.messenger
from botstory.middlewares import location
from botstory.utils import answer, SimpleTrigger

logger = logging.getLogger(__name__)


@pytest.fixture()
def mock_interface(mocker):
    mocker.patch.object(
        botstory.integrations.fb.messenger.FBInterface,
        'send_text_message',
        aiohttp.test_utils.make_mocked_coro('something'),
    )
    mocker.patch.object(
        botstory.integrations.fb.messenger.FBInterface,
        'send_list',
        aiohttp.test_utils.make_mocked_coro('something'),
    )
    mocker.patch.object(
        botstory.integrations.fb.messenger.FBInterface,
        'send_image',
        aiohttp.test_utils.make_mocked_coro('something'),
    )
    mocker.patch.object(
        botstory.integrations.fb.messenger.FBInterface,
        'send_template',
        aiohttp.test_utils.make_mocked_coro('something'),
    )
    return botstory.integrations.fb.messenger.FBInterface()


@pytest.mark.asyncio
async def test_should_say(mock_interface):
    with answer.Talk() as talk:
        story = talk.story
        story.use(mock_interface)

        @story.on('hi there!')
        def one_story():
            @story.part()
            async def then(message):
                await story.say('Nice to see you!', message['user'])

        await talk.pure_text('hi there!')

        mock_interface.send_text_message.assert_called_once_with(
            recipient=talk.user,
            text='Nice to see you!',
            options=None,
        )


# TODO: move to middlewares/location/test_location.py
@pytest.mark.asyncio
@pytest.mark.skip
async def test_ask_location(mock_interface):
    with answer.Talk() as talk:
        story = talk.story
        story.use(mock_interface)

        @story.on('SOS!')
        def one_story():
            @story.part()
            async def then(message):
                await story.ask_location('Hey, bro! Where is your rocket?', message['user'])

        await talk.pure_text('SOS!')

        mock_interface.send_text_message.assert_called_once_with(talk.user, text='Hey, bro! Where is your rocket?')


@pytest.mark.asyncio
@pytest.mark.skip
async def test_get_location_as_result_of_asking_of_location(mock_interface):
    trigger = SimpleTrigger()

    with answer.Talk() as talk:
        story = talk.story
        story.use(mock_interface)

        @story.on('SOS!')
        def one_story():
            @story.part()
            def then(ctx):
                return story.ask_location('Hey, bro! Where is your rocket?', ctx['user'])

            @story.part()
            def then(ctx):
                trigger.receive(location.get_location(ctx))

        await talk.pure_text('SOS!')
        await talk.location('somewhere')

        assert trigger.result() == 'somewhere'


@pytest.mark.asyncio
async def test_should_list_elements(mock_interface):
    with answer.Talk() as talk:
        story = talk.story
        story.use(mock_interface)

        @story.on('hi there!')
        def one_story():
            @story.part()
            async def then(ctx):
                await story.list_elements(elements=[{
                    'title': 'Classic T-Shirt Collection',  # (*) required
                    'image_url': 'https://peterssendreceiveapp.ngrok.io/img/collection.png',
                    'subtitle': 'See all our colors',
                    'default_action': {
                        'type': 'web_url',
                        'url': 'https://peterssendreceiveapp.ngrok.io/shop_collection',
                        'messenger_extensions': True,
                        'webview_height_ratio': 'tall',
                        'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                    },
                    'buttons': [{
                        'title': 'View',
                        'type': 'web_url',
                        'url': 'https://peterssendreceiveapp.ngrok.io/collection',
                        'messenger_extensions': True,
                        'webview_height_ratio': 'tall',
                        'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                    }]
                }, {
                    'title': 'Classic White T-Shirt',
                    'image_url': 'https://peterssendreceiveapp.ngrok.io/img/white-t-shirt.png',
                    'subtitle': '100% Cotton, 200% Comfortable',
                    'default_action': {
                        'type': 'web_url',
                        'url': 'https://peterssendreceiveapp.ngrok.io/view?item=100',
                        'messenger_extensions': True,
                        'webview_height_ratio': 'tall',
                        'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                    },
                    'buttons': [{
                        'title': 'Shop Now',
                        'type': 'web_url',
                        'url': 'https://peterssendreceiveapp.ngrok.io/shop?item=100',
                        'messenger_extensions': True,
                        'webview_height_ratio': 'tall',
                        'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                    }]
                }], buttons=[{
                    'title': 'View More',
                    'payload': 'payload',
                }], user=ctx['user'])

        await talk.ask('hi there!')

        mock_interface.send_list.assert_called_once_with(
            recipient=talk.user,
            elements=[{
                'title': 'Classic T-Shirt Collection',  # (*) required
                'image_url': 'https://peterssendreceiveapp.ngrok.io/img/collection.png',
                'subtitle': 'See all our colors',
                'default_action': {
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/shop_collection',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                },
                'buttons': [{
                    'title': 'View',
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/collection',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                }]
            }, {
                'title': 'Classic White T-Shirt',
                'image_url': 'https://peterssendreceiveapp.ngrok.io/img/white-t-shirt.png',
                'subtitle': '100% Cotton, 200% Comfortable',
                'default_action': {
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/view?item=100',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                },
                'buttons': [{
                    'title': 'Shop Now',
                    'type': 'web_url',
                    'url': 'https://peterssendreceiveapp.ngrok.io/shop?item=100',
                    'messenger_extensions': True,
                    'webview_height_ratio': 'tall',
                    'fallback_url': 'https://peterssendreceiveapp.ngrok.io/'
                }]
            }], buttons=[{
                'title': 'View More',
                'payload': 'payload',
            }],
            options=None,
        )


@pytest.mark.asyncio
async def test_should_send_template_based_message(mock_interface):
    with answer.Talk() as talk:
        story = talk.story
        story.use(mock_interface)

        payload = {
            'template_type': 'receipt',
            'recipient_name': 'Stephane Crozatier',
            'order_number': '12345678902',
            'currency': 'USD',
            'payment_method': 'Visa 2345',
            'order_url': 'http://petersapparel.parseapp.com/order?order_id=123456',
            'timestamp': '1428444852',
            'elements': [{
                'title': 'Classic White T-Shirt',
                'subtitle': '100% Soft and Luxurious Cotton',
                'quantity': 2,
                'price': 50,
                'currency': 'USD',
                'image_url': 'http://petersapparel.parseapp.com/img/whiteshirt.png'
            }, {
                'title': 'Classic Gray T-Shirt',
                'subtitle': '100% Soft and Luxurious Cotton',
                'quantity': 1,
                'price': 25,
                'currency': 'USD',
                'image_url': 'http://petersapparel.parseapp.com/img/grayshirt.png'
            }],
            'address': {
                'street_1': '1 Hacker Way',
                'street_2': '',
                'city': 'Menlo Park',
                'postal_code': '94025',
                'state': 'CA',
                'country': 'US'
            },
            'summary': {
                'subtotal': 75.00,
                'shipping_cost': 4.95,
                'total_tax': 6.19,
                'total_cost': 56.14
            },
            'adjustments': [{
                'name': 'New Customer Discount',
                'amount': 20
            }, {
                'name': '$10 Off Coupon',
                'amount': 10
            }]
        }

        await story.send_template(payload,
                                  user=talk.user)

        mock_interface.send_template.assert_called_once_with(recipient=talk.user,
                                                             payload=payload)


@pytest.mark.asyncio
async def test_send_image(mock_interface):
    with answer.Talk() as talk:
        story = talk.story
        story.use(mock_interface)

        await story.send_image('http://some.ua/image.gif',
                               user=talk.user)

        mock_interface.send_image.assert_called_once_with(talk.user, 'http://some.ua/image.gif', None)



import asyncio
from botstory.ast import story_context
from botstory.integrations.commonhttp import errors as commonhttp_errors
import functools
import logging
from . import validate
from .. import commonhttp
from ... import di
from ...middlewares import option

logger = logging.getLogger(__name__)


@di.desc('fb', reg=False)
class FBInterface:
    type = 'facebook'

    def __init__(self,
                 api_uri='https://graph.facebook.com/v2.6',
                 greeting_text=None,
                 page_access_token='?',
                 persistent_menu=None,
                 webhook_url=None,
                 webhook_token=None,
                 ):
        """

        :param api_uri:
        :param greeting_text:
        :param page_access_token:
        :param persistent_menu:
        :param webhook_url:
        :param webhook_token:
        """
        self.api_uri = api_uri
        self.greeting_text = greeting_text
        self.persistent_menu = persistent_menu
        self.token = page_access_token
        self.webhook = webhook_url
        self.webhook_token = webhook_token

        self.library = None
        self.http = None
        self.story_processor = None
        self.storage = None
        self.users = None

    @di.inject()
    def add_library(self, stories_library):
        logger.debug('add_library')
        logger.debug(stories_library)
        self.library = stories_library

    @di.inject()
    def add_http(self, http):
        """
        inject http provider

        :param http:
        :return:
        """
        logger.debug('add_http')
        logger.debug(http)
        self.http = http

    @di.inject()
    def add_processor(self, story_processor):
        logger.debug('add_processor')
        logger.debug(story_processor)
        self.story_processor = story_processor

    @di.inject()
    def add_storage(self, storage):
        logger.debug('add_storage')
        logger.debug(storage)
        self.storage = storage

    @di.inject()
    def add_users(self, users):
        logger.debug('add_users')
        logger.debug(users)
        self.users = users

    async def send_text_message(self, recipient, text,
                                quick_replies=None,
                                options=None):
        """
        async send message to the facebook user (recipient)

        :param recipient:
        :param text:
        :param quick_replies:
        :param options

        :return:
        """

        try:
            validate.send_text_message(text, quick_replies)
        except validate.ExceedLengthException as i:
            # TODO: take first part of message show option `more`
            # store last part until user press `more`

            # TODO: register dynamic `quick answer` handler
            # with the rest of message

            # TODO: or handle it on application level?
            # motivation: if we're working with dynamic data or endless
            # we could just pass text. It should be generator
            logger.warn(str(i))

            if options and options.get('overflow', None) == 'cut':
                text = text[:i.limit]
            else:
                text = text[:i.limit - 2] + '\u2026'
        except validate.Invalid as i:
            logger.warn(str(i))

        if not quick_replies:
            quick_replies = []

        message = {
            'text': text,
        }

        quick_replies = [{'content_type': 'text', **reply} for reply in quick_replies]
        if len(quick_replies) > 0:
            message['quick_replies'] = quick_replies

        return await self.http.post(
            self.api_uri + '/me/messages/',
            params={
                'access_token': self.token,
            },
            json={
                'recipient': {
                    'id': recipient['facebook_user_id'],
                },
                'message': message,
            })

    async def send_list(self, recipient, elements, buttons=None, options=None):
        """
        send list of elements
        :param recipient:
        :param elements:
        :param buttons:
        :return:
        """
        # try:
        #     validate.send_listd
        # except
        # ....

        buttons = [{**b, 'type': b.get('type', 'postback')} for b in buttons]

        options = options or {}

        return await self.send_template(recipient, {
            'template_type': 'list',
            'top_element_style': options.get('top_element_style', 'large'),
            'elements': elements,
            'buttons': buttons,
        })

    async def send_image(self, recipient, url, options=None):
        if options is None:
            options = {}

        should_try = True
        delay = options.get('retry_delay', 1)
        tries = options.get('retry_times', 3)

        while should_try:
            try:
                await self._send_image(recipient, url)
                should_try = False
            except commonhttp_errors.HttpRequestError as err:
                if tries > 0:
                    tries -= 1
                    logger.warning('# retry to send image {}'.format(url))
                    should_try = True
                    await asyncio.sleep(delay)
                else:
                    raise err

    async def _send_image(self, recipient, url):
        return await self.http.post(
            self.api_uri + '/me/messages/',
            params={
                'access_token': self.token,
            },
            json={
                'recipient': {
                    'id': recipient['facebook_user_id'],
                },

                'message': {
                    'attachment': {
                        'type': 'image',
                        'payload': {
                            'url': url,
                        },
                    },
                },
            })

    async def send_template(self, recipient, payload):
        """
        send template based message

        :param recipient:
        :param payload:
        :return:
        """
        return await self.http.post(
            self.api_uri + '/me/messages/',
            params={
                'access_token': self.token,
            },
            json={
                'recipient': {
                    'id': recipient['facebook_user_id'],
                },

                'message': {
                    'attachment': {
                        'type': 'template',
                        'payload': payload,
                    }
                },
            })

    async def request_profile(self, facebook_user_id):
        """
        Make request to facebook
        to receive more information about user

        More: https://developers.facebook.com/docs/messenger-platform/user-profile

        :param facebook_user_id:
        :return:
        """
        return await self.http.get(
            '{}/{}'.format(self.api_uri, facebook_user_id),
            params={
                'access_token': self.token,
            },
        )

    async def handle(self, data):
        """
        handle fb messenger message and return 200Ok as quick as possible
        In background we launch process of handling incoming message
        This solution is ok only for small apps for bigger one it is much better
        to use a queue.

        TODO: support third party queue

        :param data:
        :return:
        """
        loop = asyncio.get_event_loop()
        loop.create_task(self.process(data))
        return {
            'status': 200,
            'text': 'Ok!',
        }

    async def process(self, data):
        """
        async process message which comes from fb messenger

        :param data:
        :return:
        """
        logger.debug('')
        logger.debug('> handle <')
        logger.debug('')
        logger.debug('  entry: {}'.format(data))
        try:
            for e in data.get('entry', []):
                messaging = e.get('messaging', [])
                logger.debug('  messaging: {}'.format(messaging))

                if len(messaging) == 0:
                    logger.warning('  entry {} list lack of "messaging" field'.format(e))

                for m in messaging:
                    logger.debug('  m: {}'.format(m))

                    facebook_user_id = m['sender']['id']

                    logger.debug('before get user with facebook_user_id={}'.format(facebook_user_id))
                    user = await self.storage.get_user(facebook_user_id=facebook_user_id)
                    if not user:
                        logger.debug('  should create new user {}'.format(facebook_user_id))

                        try:
                            messenger_profile_data = await self.request_profile(facebook_user_id)
                            logger.debug('receive fb profile {}'.format(messenger_profile_data))
                        except commonhttp.errors.HttpRequestError as err:
                            logger.debug('fail on request fb profile of {}. with {}'.format(facebook_user_id, err))
                            messenger_profile_data = {
                                'no_fb_profile': True,
                            }

                        logger.debug('before creating new user')
                        user = await self.storage.new_user(
                            facebook_user_id=facebook_user_id,
                            no_fb_profile=messenger_profile_data.get('no_fb_profile', None),
                            first_name=messenger_profile_data.get('first_name', None),
                            last_name=messenger_profile_data.get('last_name', None),
                            profile_pic=messenger_profile_data.get('profile_pic', None),
                            locale=messenger_profile_data.get('locale', None),
                            timezone=messenger_profile_data.get('timezone', None),
                            gender=messenger_profile_data.get('gender', None),
                        )

                        self.users.on_new_user_comes(user)

                    session = await self.storage.get_session(facebook_user_id=facebook_user_id)
                    if not session:
                        logger.debug('  should create new session for user {}'.format(facebook_user_id))
                        session = await self.storage.new_session(
                            facebook_user_id=facebook_user_id,
                            stack=[],
                            user=user,
                        )

                    ctx = story_context.clean_message_data({
                        'session': session,
                        'user': user,
                    })

                    if 'message' in m:
                        logger.debug('message notification')
                        raw_message = m.get('message', {})
                        if 'is_echo' in raw_message:
                            # TODO: should react somehow.
                            # for example storing for debug purpose
                            logger.debug('just echo message')
                        else:
                            text = raw_message.get('text', None)
                            if text is not None:
                                ctx = story_context.set_message_data(ctx,
                                                                     'text', {
                                                                         'raw': text,
                                                                     })
                            elif 'sticker_id' in raw_message:
                                ctx = story_context.set_message_data(ctx,
                                                                     'sticker_id', raw_message['sticker_id'],
                                                                     )
                            elif 'attachments' in raw_message:
                                ctx = story_context.set_message_data(ctx,
                                                                     'attachments', raw_message['attachments'])
                            else:
                                logger.warning('  entry {} "text"'.format(e))

                            quick_reply = raw_message.get('quick_reply', None)
                            if quick_reply is not None:
                                ctx = story_context.set_message_data(ctx,
                                                                     'option',
                                                                     'value', quick_reply['payload'])

                            ctx = await self.story_processor.match_message(ctx)

                    elif 'postback' in m:
                        ctx = story_context.set_message_data(ctx,
                                                             'option',
                                                             'value', m['postback']['payload'])
                        ctx = await self.story_processor.match_message(ctx)
                    elif 'delivery' in m:
                        logger.debug('delivery notification')
                    elif 'read' in m:
                        logger.debug('read notification')
                    else:
                        logger.warning('(!) unknown case {}'.format(e))

                    # after message were processed session and user information could change
                    # so we should store it for the next usage
                    await self.storage.set_session(ctx['session'])
                    await self.storage.set_user(ctx['user'])

        except BaseException as err:
            logger.exception(err)

        return {
            'status': 200,
            'text': 'Ok!',
        }

    async def setup(self):
        logger.debug('setup')

        if self.greeting_text:
            await self.replace_greeting_text(self.greeting_text)

        # check whether we have `On Start Story`

        # error if we don't have start button handler but have persistent menu:
        # (#100) You must set a Get Started button if you also wish to use persistent menu.
        have_on_start_story = not not self.library.get_global_story(story_context.set_message_data({
            'session': {}
        }, 'option',
            'value', option.OnStart.DEFAULT_OPTION_PAYLOAD))

        if have_on_start_story or self.persistent_menu:
            await self.remove_greeting_call_to_action_payload()
            await self.set_greeting_call_to_action_payload(option.OnStart.DEFAULT_OPTION_PAYLOAD)

        if self.persistent_menu:
            await self.replace_persistent_menu(self.persistent_menu)

        await self.subscribe()

    async def before_start(self):
        logger.debug('# start')
        if self.webhook and self.http:
            self.http.webhook(self.webhook, self.handle, self.webhook_token)

    async def replace_greeting_text(self, message):
        """
        delete greeting text before
        :param message:
        :return:
        """
        try:
            await self.remove_greeting_text()
        except Exception:
            pass

        await self.set_greeting_text(message)

    async def set_greeting_text(self, message):
        """
        set a greeting for new conversations

        can use for personalizing
            {{user_first_name}}
            {{user_last_name}}
            {{user_full_name}}

        more: https://developers.facebook.com/docs/messenger-platform/thread-settings/greeting-text

        :param message:
        :return:
        """
        logger.debug('set_greeting_text')

        greeting_data = {
            'locale': 'default',
            'text': message,
        }

        try:
            validate.greeting_text(message)
        except validate.Invalid as i:
            logger.warn(str(i))

        self.greeting_text = message

        if not self.http:
            # should wait until receive http
            return

        await self.http.post(
            self.api_uri + '/me/messenger_profile',
            params={
                'access_token': self.token,
            },
            json={
                'greeting': [
                    greeting_data,
                ],
            }
        )

    async def remove_greeting_text(self):
        logger.debug('# remove_greeting_text')
        if not self.http:
            return

        await self.http.delete(
            self.api_uri + '/me/messenger_profile',
            params={
                'access_token': self.token,
            },
            json={
                'fields': [
                    'greeting',
                ]
            }
        )

    async def set_greeting_call_to_action_payload(self, payload):
        logger.debug('# set_greeting_call_to_action_payload')
        """

        more: https://developers.facebook.com/docs/messenger-platform/thread-settings/get-started-button

        :param payload:
        :return:
        """
        await self.http.post(
            self.api_uri + '/me/messenger_profile',
            params={
                'access_token': self.token,
            },
            json={
                'get_started': {'payload': payload}
            }
        )

    async def remove_greeting_call_to_action_payload(self):
        logger.debug('# remove_greeting_call_to_action_payload')
        if not self.http:
            return

        await self.http.delete(
            self.api_uri + '/me/messenger_profile',
            params={
                'access_token': self.token,
            },
            json={
                'fields': [
                    'get_started',
                ]
            }
        )

    async def replace_persistent_menu(self, menu):
        logger.debug('# replace_persistent_menu')
        try:
            await self.remove_persistent_menu()
        except Exception:
            pass
        await self.set_persistent_menu(menu)

    async def set_persistent_menu(self, menu):
        """
        more: https://developers.facebook.com/docs/messenger-platform/thread-settings/persistent-menu

        :param menu:
        :return:
        """
        logger.debug('# set_persistent_menu')
        if not isinstance(menu, list) or len(menu) == 0:
            raise ValueError(
                'Menu should be not empty list. '
                'More info here: '
                'https://developers.facebook.com/docs/messenger-platform/messenger-profile/persistent-menu')
        if all('locale' in item for item in menu):
            # TODO: should validate as well
            persistent_menu = menu
        else:
            try:
                validate.persistent_menu(menu)
            except validate.Invalid as i:
                logger.warn(str(i))
            persistent_menu = [{
                'locale': 'default',
                'call_to_actions': menu,
            }]

        self.persistent_menu = persistent_menu

        if not self.http:
            # should wait until receive http
            logger.warning('There is no http interface yet')
            return

        await self.http.post(
            self.api_uri + '/me/messenger_profile',
            params={
                'access_token': self.token,
            },
            json={
                'persistent_menu': persistent_menu,
            }
        )

    async def remove_persistent_menu(self):
        logger.debug('# remove_persistent_menu')
        if not self.http:
            return

        await self.http.delete(
            self.api_uri + '/me/messenger_profile',
            params={
                'access_token': self.token,
            },
            json={'fields': [
                'persistent_menu',
            ]}
        )

    async def start_typing(self, user):
        logger.debug('# start typing')
        await self.http.post(
            self.api_uri + '/me/messages',
            params={
                'access_token': self.token,
            },
            json={
                'recipient': {
                    'id': user['facebook_user_id'],
                },
                'sender_action': 'typing_on',
            },
        )

    async def stop_typing(self, user):
        logger.debug('# stop typing')
        await self.http.post(
            self.api_uri + '/me/messages',
            params={
                'access_token': self.token,
            },
            json={
                'recipient': {
                    'id': user['facebook_user_id'],
                },
                'sender_action': 'typing_off',
            },
        )

    async def subscribe(self):
        """
        subscribe the app to page updates
        docs:
        https://developers.facebook.com/docs/messenger-platform/guides/setup#subscribe_app

        :return:
        """
        logger.debug('# subscribe page')
        await self.http.post(
            self.api_uri + '/me/subscribed_apps',
            params={
                'access_token': self.token,
            },
        )



from .ast.callable import EndOfStory
from .ast.forking import SwitchOnValue, Undefined
from .story import Story
import os

__all__ = [EndOfStory, Story, SwitchOnValue]

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'version.txt')) as version_file:
    __version__ = version_file.read().strip()
